@@ -1,15 +1,23 @@
-from abc import (abstractmethod, ABC)
+from abc import abstractmethod, ABC
 from collections import OrderedDict
-import os
 from string import Formatter
+from typing import Any, Callable, Dict, List, Optional, TYPE_CHECKING, Union
 
 import pystac
+from pystac.utils import safe_urlparse, join_path_or_url, JoinType
+
+if TYPE_CHECKING:
+    from pystac.stac_object import STACObject as STACObject_Type
+    from pystac.catalog import Catalog as Catalog_Type
+    from pystac.collection import Collection as Collection_Type
+    from pystac.item import Item as Item_Type
 
 
 class TemplateError(Exception):
     """Exception thrown when an error occurs during converting a template
     string into data for :class:`~pystac.layout.LayoutTemplate`
     """
+
     pass
 
 
@@ -58,81 +66,104 @@ class LayoutTemplate:
         template = LayoutTemplate("${collection}/${common_metadata.license}")
 
     Args:
-        template (str): The template string to use.
-        defaults (dict): A dictionary of template vars to values. These values
+        template : The template string to use.
+        defaults : A dictionary of template vars to values. These values
             will be used in case a value cannot be derived from a stac object.
     """
 
     # Special template vars specific to Items
-    ITEM_TEMPLATE_VARS = ['date', 'year', 'month', 'day', 'collection']
+    ITEM_TEMPLATE_VARS = ["date", "year", "month", "day", "collection"]
 
-    def __init__(self, template, defaults=None):
+    def __init__(
+        self, template: str, defaults: Optional[Dict[str, str]] = None
+    ) -> None:
         self.template = template
         self.defaults = defaults or {}
 
         # Generate list of template vars
-        template_vars = []
+        template_vars: List[str] = []
         for formatter_parse_result in Formatter().parse(template):
             v = formatter_parse_result[1]
             if v is not None:
-                if formatter_parse_result[2] != '':
-                    v = '{}:{}'.format(v, formatter_parse_result[2])
+                if formatter_parse_result[2] != "":
+                    v = "{}:{}".format(v, formatter_parse_result[2])
                 template_vars.append(v)
         self.template_vars = template_vars
 
-    def _get_template_value(self, stac_object, template_var):
+    def _get_template_value(
+        self, stac_object: "STACObject_Type", template_var: str
+    ) -> Any:
         if template_var in self.ITEM_TEMPLATE_VARS:
-            if stac_object.STAC_OBJECT_TYPE == pystac.STACObjectType.ITEM:
+            if isinstance(stac_object, pystac.Item):
                 # Datetime
                 dt = stac_object.datetime
                 if dt is None:
                     dt = stac_object.common_metadata.start_datetime
                 if dt is None:
-                    raise TemplateError('Item {} does not have a datetime or '
-                                        'datetime range set; cannot template {} in {}'.format(
-                                            stac_object, template_var, self.template))
+                    raise TemplateError(
+                        "Item {} does not have a datetime or "
+                        "datetime range set; cannot template {} in {}".format(
+                            stac_object, template_var, self.template
+                        )
+                    )
 
-                if template_var == 'year':
+                if template_var == "year":
                     return dt.year
-                if template_var == 'month':
+                if template_var == "month":
                     return dt.month
-                if template_var == 'day':
+                if template_var == "day":
                     return dt.day
-                if template_var == 'date':
+                if template_var == "date":
                     return dt.date().isoformat()
 
                 # Collection
-                if template_var == 'collection':
+                if template_var == "collection":
                     if stac_object.collection_id is not None:
                         return stac_object.collection_id
                     raise TemplateError(
-                        'Item {} does not have a collection ID set; cannot template {} in {}'.
-                        format(stac_object, template_var, self.template))
+                        f"Item {stac_object} does not have a collection ID set; "
+                        f"cannot template {template_var} in {self.template}"
+                    )
             else:
-                raise TemplateError('"{}" cannot be used to template non-Item {} in {}'.format(
-                    template_var, stac_object, self.template))
+                raise TemplateError(
+                    '"{}" cannot be used to template non-Item {} in {}'.format(
+                        template_var, stac_object, self.template
+                    )
+                )
 
         # Allow dot-notation properties for arbitrary object values.
-        props = template_var.split('.')
-        prop_location = None
-        error = TemplateError('Cannot find property {} on {} for template {}'.format(
-            template_var, stac_object, self.template))
+        props = template_var.split(".")
+        prop_source: Optional[Union[pystac.STACObject, Dict[str, Any]]] = None
+        error = TemplateError(
+            "Cannot find property {} on {} for template {}".format(
+                template_var, stac_object, self.template
+            )
+        )
 
         try:
+
             if hasattr(stac_object, props[0]):
-                prop_location = stac_object
-            elif hasattr(
-                    stac_object,
-                    "properties") and stac_object.properties and props[0] in stac_object.properties:
-                prop_location = stac_object.properties
-            elif hasattr(stac_object, "extra_fields"
-                         ) and stac_object.extra_fields and props[0] in stac_object.extra_fields:
-                prop_location = stac_object.extra_fields
-            else:
+                prop_source = stac_object
+
+            if prop_source is None and hasattr(stac_object, "properties"):
+                obj_props: Optional[
+                    Dict[str, Any]
+                ] = stac_object.properties  # type:ignore
+                if obj_props is not None and props[0] in obj_props:
+                    prop_source = obj_props
+
+            if prop_source is None and hasattr(stac_object, "extra_fields"):
+                extra_fields: Optional[
+                    Dict[str, Any]
+                ] = stac_object.extra_fields  # type:ignore
+                if extra_fields is not None and props[0] in extra_fields:
+                    prop_source = extra_fields
+
+            if prop_source is None:
                 raise error
 
-            v = prop_location
-            for prop in template_var.split('.'):
+            v: Any = prop_source
+            for prop in template_var.split("."):
                 if type(v) is dict:
                     if prop not in v:
                         raise error
@@ -148,14 +179,14 @@ class LayoutTemplate:
 
         return v
 
-    def get_template_values(self, stac_object):
+    def get_template_values(self, stac_object: "STACObject_Type") -> Dict[str, Any]:
         """Gets a dictionary of template variables to values derived from
         the given stac_object. If the template vars cannot be found in the
         stac object, and defaults was supplied to this template, a default
         value is used; if there is no default then this will raise an error.
 
         Args:
-            stac_object (STACObject): The STACObject to derive template
+            stac_object : The STACObject to derive template
                 variable values from.
 
         Returns:
@@ -168,20 +199,21 @@ class LayoutTemplate:
                 derived from the stac object and there is no default,
                 this error will be raised.
         """
-        return OrderedDict([(k, self._get_template_value(stac_object, k))
-                            for k in self.template_vars])
+        return OrderedDict(
+            [(k, self._get_template_value(stac_object, k)) for k in self.template_vars]
+        )
 
-    def substitute(self, stac_object):
+    def substitute(self, stac_object: "STACObject_Type") -> str:
         """Substitutes the values derived from
         :meth:`~pystac.layout.LayoutTemplate.get_template_values` into
         the template string for this template.
 
         Args:
-            stac_object (STACObject): The STACObject to derive template
+            stac_object : The STACObject to derive template
                 variable values from.
 
         Returns:
-            [str]: The original template supplied to this LayoutTemplate
+            str: The original template supplied to this LayoutTemplate
                 with template variables replaced by the values derived
                 from this stac object.
 
@@ -194,34 +226,40 @@ class LayoutTemplate:
 
         s = self.template
         for key, value in parts.items():
-            s = s.replace('${' + '{}'.format(key) + '}', '{}'.format(value))
+            s = s.replace("${" + "{}".format(key) + "}", "{}".format(value))
         return s
 
 
 class HrefLayoutStrategy(ABC):
     """Base class for HREF Layout strategies."""
-    def get_href(self, stac_object, parent_dir, is_root=False):
-        stac_object_type = stac_object.STAC_OBJECT_TYPE
-        if stac_object_type == pystac.STACObjectType.CATALOG:
-            return self.get_catalog_href(stac_object, parent_dir, is_root)
-        elif stac_object_type == pystac.STACObjectType.COLLECTION:
-            return self.get_collection_href(stac_object, parent_dir, is_root)
-        elif stac_object_type == pystac.STACObjectType.ITEM:
+
+    def get_href(
+        self, stac_object: "STACObject_Type", parent_dir: str, is_root: bool = False
+    ) -> str:
+        if isinstance(stac_object, pystac.Item):
             return self.get_item_href(stac_object, parent_dir)
+        elif isinstance(stac_object, pystac.Collection):
+            return self.get_collection_href(stac_object, parent_dir, is_root)
+        elif isinstance(stac_object, pystac.Catalog):
+            return self.get_catalog_href(stac_object, parent_dir, is_root)
         else:
-            raise pystac.STACError('Unknown STAC object type {}'.format(stac_object_type))
+            raise pystac.STACError("Unknown STAC object type {}".format(stac_object))
 
     @abstractmethod
-    def get_catalog_href(self, cat, parent_dir, is_root):
-        pass
+    def get_catalog_href(
+        self, cat: "Catalog_Type", parent_dir: str, is_root: bool
+    ) -> str:
+        raise NotImplementedError
 
     @abstractmethod
-    def get_collection_href(self, col, parent_dir, is_root):
-        pass
+    def get_collection_href(
+        self, col: "Collection_Type", parent_dir: str, is_root: bool
+    ) -> str:
+        raise NotImplementedError
 
     @abstractmethod
-    def get_item_href(self, item, parent_dir):
-        pass
+    def get_item_href(self, item: "Item_Type", parent_dir: str) -> str:
+        raise NotImplementedError
 
 
 class CustomLayoutStrategy(HrefLayoutStrategy):
@@ -229,48 +267,55 @@ class CustomLayoutStrategy(HrefLayoutStrategy):
     stac object paths.
 
     Args:
-        catalog_func (Callable[Catalog, str, bool] -> str): A function that takes
+        catalog_func : A function that takes
             an catalog, a parent directory, and a flag specifying whether
             or not this catalog is the root. If it is the root, its usually
             best to not create a subdirectory and put the Catalog file directly
             in the parent directory. Must return the string path.
-        collection_func (Callable[Catalog, str, bool] -> str): A function that
+        collection_func : A function that
             is used for collections in the same manner at ``catalog_func``. This
             takes the same parameters.
-        item_func  (Callable[Catalog, str] -> str): A function that takes
+        item_func  : A function that takes
             an item and a parent directory and returns the path to be used
             for the item.
-        fallback_strategy (HrefLayoutStrategy): The fallback strategy to
+        fallback_strategy : The fallback strategy to
             use if a function is not provided for a stac object type. Defaults to
             :class:`~pystac.layout.BestPracticesLayoutStrategy`
     """
-    def __init__(self,
-                 catalog_func=None,
-                 collection_func=None,
-                 item_func=None,
-                 fallback_strategy=None):
+
+    def __init__(
+        self,
+        catalog_func: Optional[Callable[["Catalog_Type", str, bool], str]] = None,
+        collection_func: Optional[Callable[["Collection_Type", str, bool], str]] = None,
+        item_func: Optional[Callable[["Item_Type", str], str]] = None,
+        fallback_strategy: Optional[HrefLayoutStrategy] = None,
+    ):
         self.item_func = item_func
         self.collection_func = collection_func
         self.catalog_func = catalog_func
         if fallback_strategy is None:
             fallback_strategy = BestPracticesLayoutStrategy()
-        self.fallback_strategy = fallback_strategy
+        self.fallback_strategy: HrefLayoutStrategy = fallback_strategy
 
-    def get_catalog_href(self, cat, parent_dir, is_root):
+    def get_catalog_href(
+        self, cat: "Catalog_Type", parent_dir: str, is_root: bool
+    ) -> str:
         if self.catalog_func is not None:
             result = self.catalog_func(cat, parent_dir, is_root)
             if result is not None:
                 return result
         return self.fallback_strategy.get_catalog_href(cat, parent_dir, is_root)
 
-    def get_collection_href(self, col, parent_dir, is_root):
+    def get_collection_href(
+        self, col: "Collection_Type", parent_dir: str, is_root: bool
+    ) -> str:
         if self.collection_func is not None:
             result = self.collection_func(col, parent_dir, is_root)
             if result is not None:
                 return result
         return self.fallback_strategy.get_collection_href(col, parent_dir, is_root)
 
-    def get_item_href(self, item, parent_dir):
+    def get_item_href(self, item: "Item_Type", parent_dir: str) -> str:
         if self.item_func is not None:
             result = self.item_func(item, parent_dir)
             if result is not None:
@@ -289,69 +334,97 @@ class TemplateLayoutStrategy(HrefLayoutStrategy):
     object.
 
     Args:
-        catalog_template (str): The template string to use for catalog paths.
+        catalog_template : The template string to use for catalog paths.
             Must be a valid template string that can be used by
             :class:`~pystac.layout.LayoutTemplate`
-        collection_template (str): The template string to use for collection paths.
+        collection_template : The template string to use for collection paths.
             Must be a valid template string that can be used by
             :class:`~pystac.layout.LayoutTemplate`
-        item_template (str): The template string to use for item paths.
+        item_template : The template string to use for item paths.
             Must be a valid template string that can be used by
             :class:`~pystac.layout.LayoutTemplate`
-        fallback_strategy (HrefLayoutStrategy): The fallback strategy to
+        fallback_strategy : The fallback strategy to
             use if a template is not provided. Defaults to
             :class:`~pystac.layout.BestPracticesLayoutStrategy`
     """
-    def __init__(self,
-                 catalog_template=None,
-                 collection_template=None,
-                 item_template=None,
-                 fallback_strategy=None):
-        self.catalog_template = LayoutTemplate(
-            catalog_template) if catalog_template is not None else None
-        self.collection_template = LayoutTemplate(
-            collection_template) if collection_template is not None else None
-        self.item_template = LayoutTemplate(item_template) if item_template is not None else None
+
+    def __init__(
+        self,
+        catalog_template: Optional[str] = None,
+        collection_template: Optional[str] = None,
+        item_template: Optional[str] = None,
+        fallback_strategy: Optional[HrefLayoutStrategy] = None,
+    ):
+        self.catalog_template = (
+            LayoutTemplate(catalog_template) if catalog_template is not None else None
+        )
+        self.collection_template = (
+            LayoutTemplate(collection_template)
+            if collection_template is not None
+            else None
+        )
+        self.item_template = (
+            LayoutTemplate(item_template) if item_template is not None else None
+        )
 
         if fallback_strategy is None:
             fallback_strategy = BestPracticesLayoutStrategy()
-        self.fallback_strategy = fallback_strategy
+        self.fallback_strategy: HrefLayoutStrategy = fallback_strategy
+
+    def get_catalog_href(
+        self, cat: "Catalog_Type", parent_dir: str, is_root: bool
+    ) -> str:
+        parsed_parent_dir = safe_urlparse(parent_dir)
+        join_type = JoinType.from_parsed_uri(parsed_parent_dir)
 
-    def get_catalog_href(self, cat, parent_dir, is_root):
         if is_root or self.catalog_template is None:
             return self.fallback_strategy.get_catalog_href(cat, parent_dir, is_root)
         else:
             template_path = self.catalog_template.substitute(cat)
-            if not template_path.endswith('.json'):
-                template_path = os.path.join(template_path, cat.DEFAULT_FILE_NAME)
+            if not template_path.endswith(".json"):
+                template_path = join_path_or_url(
+                    join_type, template_path, cat.DEFAULT_FILE_NAME
+                )
 
-            return os.path.join(parent_dir, template_path)
+            return join_path_or_url(join_type, parent_dir, template_path)
+
+    def get_collection_href(
+        self, col: "Collection_Type", parent_dir: str, is_root: bool
+    ) -> str:
+        parsed_parent_dir = safe_urlparse(parent_dir)
+        join_type = JoinType.from_parsed_uri(parsed_parent_dir)
 
-    def get_collection_href(self, col, parent_dir, is_root):
         if is_root or self.collection_template is None:
             return self.fallback_strategy.get_collection_href(col, parent_dir, is_root)
         else:
             template_path = self.collection_template.substitute(col)
-            if not template_path.endswith('.json'):
-                template_path = os.path.join(template_path, col.DEFAULT_FILE_NAME)
+            if not template_path.endswith(".json"):
+                template_path = join_path_or_url(
+                    join_type, template_path, col.DEFAULT_FILE_NAME
+                )
 
-            return os.path.join(parent_dir, template_path)
+            return join_path_or_url(join_type, parent_dir, template_path)
+
+    def get_item_href(self, item: "Item_Type", parent_dir: str) -> str:
+        parsed_parent_dir = safe_urlparse(parent_dir)
+        join_type = JoinType.from_parsed_uri(parsed_parent_dir)
 
-    def get_item_href(self, item, parent_dir):
         if self.item_template is None:
             return self.fallback_strategy.get_item_href(item, parent_dir)
         else:
             template_path = self.item_template.substitute(item)
-            if not template_path.endswith('.json'):
-                template_path = os.path.join(template_path, '{}.json'.format(item.id))
+            if not template_path.endswith(".json"):
+                template_path = join_path_or_url(
+                    join_type, template_path, "{}.json".format(item.id)
+                )
 
-            return os.path.join(parent_dir, template_path)
+            return join_path_or_url(join_type, parent_dir, template_path)
 
 
 class BestPracticesLayoutStrategy(HrefLayoutStrategy):
     """Layout strategy that represents the catalog layout described
-    in the `STAC Best Practices documentation
-    <https://github.com/radiantearth/stac-spec/blob/v1.0.0-beta.2/best-practices.md>`_
+    in the :stac-spec:`STAC Best Practices documentation
+    <best-practices.md>`
 
     For a root catalog or collection, this will use the filename 'catalog.json'
     or 'collection.json' to the given directory. For a non-root catalog or collection,
@@ -362,23 +435,37 @@ class BestPracticesLayoutStrategy(HrefLayoutStrategy):
 
     All paths are appended to the parent directory.
     """
-    def get_catalog_href(self, cat, parent_dir, is_root):
+
+    def get_catalog_href(
+        self, cat: "Catalog_Type", parent_dir: str, is_root: bool
+    ) -> str:
+        parsed_parent_dir = safe_urlparse(parent_dir)
+        join_type = JoinType.from_parsed_uri(parsed_parent_dir)
+
         if is_root:
             cat_root = parent_dir
         else:
-            cat_root = os.path.join(parent_dir, '{}'.format(cat.id))
+            cat_root = join_path_or_url(join_type, parent_dir, "{}".format(cat.id))
 
-        return os.path.join(cat_root, cat.DEFAULT_FILE_NAME)
+        return join_path_or_url(join_type, cat_root, cat.DEFAULT_FILE_NAME)
+
+    def get_collection_href(
+        self, col: "Collection_Type", parent_dir: str, is_root: bool
+    ) -> str:
+        parsed_parent_dir = safe_urlparse(parent_dir)
+        join_type = JoinType.from_parsed_uri(parsed_parent_dir)
 
-    def get_collection_href(self, col, parent_dir, is_root):
         if is_root:
             col_root = parent_dir
         else:
-            col_root = os.path.join(parent_dir, '{}'.format(col.id))
+            col_root = join_path_or_url(join_type, parent_dir, "{}".format(col.id))
 
-        return os.path.join(col_root, col.DEFAULT_FILE_NAME)
+        return join_path_or_url(join_type, col_root, col.DEFAULT_FILE_NAME)
 
-    def get_item_href(self, item, parent_dir):
-        item_root = os.path.join(parent_dir, '{}'.format(item.id))
+    def get_item_href(self, item: "Item_Type", parent_dir: str) -> str:
+        parsed_parent_dir = safe_urlparse(parent_dir)
+        join_type = JoinType.from_parsed_uri(parsed_parent_dir)
 
-        return os.path.join(item_root, '{}.json'.format(item.id))
+        item_root = join_path_or_url(join_type, parent_dir, "{}".format(item.id))
+
+        return join_path_or_url(join_type, item_root, "{}.json".format(item.id))
