@@ -1,20 +1,21 @@
 import os
+from typing import Optional
 
-__version__ = '0.5.4'
+__version__ = "1.1.0"
 """Library version"""
 
 
 class STACVersion:
-    DEFAULT_STAC_VERSION = '1.0.0-beta.2'
+    DEFAULT_STAC_VERSION = "1.0.0"
     """Latest STAC version supported by PySTAC"""
 
     # Version that holds a user-set STAC version to use.
-    _override_version = None
+    _override_version: Optional[str] = None
 
-    OVERRIDE_VERSION_ENV_VAR = 'PYSTAC_STAC_VERSION_OVERRIDE'
+    OVERRIDE_VERSION_ENV_VAR = "PYSTAC_STAC_VERSION_OVERRIDE"
 
     @classmethod
-    def get_stac_version(cls):
+    def get_stac_version(cls) -> str:
         if cls._override_version is not None:
             return cls._override_version
 
@@ -25,17 +26,18 @@ class STACVersion:
         return cls.DEFAULT_STAC_VERSION
 
     @classmethod
-    def set_stac_version(cls, stac_version):
+    def set_stac_version(cls, stac_version: Optional[str]) -> None:
         cls._override_version = stac_version
 
 
-def get_stac_version():
+def get_stac_version() -> str:
     """Returns the STAC version PySTAC writes as the "stac_version" property for
     any object it serializes into JSON.
 
-    If a call to ``set_stac_version`` was made, this will return the value it was called with.
-    Next it will check the environment for a PYSTAC_STAC_VERSION_OVERRIDE variable. Otherwise
-    it will return the latest STAC version that this version of PySTAC supports.
+    If a call to ``set_stac_version`` was made, this will return the value it was
+    called with. Next it will check the environment for a PYSTAC_STAC_VERSION_OVERRIDE
+    variable. Otherwise it will return the latest STAC version that this version of
+    PySTAC supports.
 
     Returns:
         str: The STAC Version PySTAC is set up to use.
@@ -43,7 +45,7 @@ def get_stac_version():
     return STACVersion.get_stac_version()
 
 
-def set_stac_version(stac_version):
+def set_stac_version(stac_version: Optional[str]) -> None:
     """Sets the STAC version that PySTAC should use.
 
     This is the version that will be set as the "stac_version" property
@@ -55,12 +57,14 @@ def set_stac_version(stac_version):
     the version.
 
     Args:
-        stac_version (str): The STAC version to use instead of the latest STAC version that
-            PySTAC supports (described in STACVersion.DEFAULT_STAC_VERSION)
+        stac_version : The STAC version to use instead of the latest STAC version
+            that PySTAC supports (described in STACVersion.DEFAULT_STAC_VERSION).
+            If None, clear to use the default for this version of PySTAC.
 
     Note:
-        Setting the STAC version to something besides the default version will not effect
-        the format of STAC read or written; it will only override the ``stac_version`` property
-        of the objects being written. Setting this incorrectly can produce invalid STAC.
+        Setting the STAC version to something besides the default version will not
+        effect the format of STAC read or written; it will only override the
+        ``stac_version`` property of the objects being written. Setting this
+        incorrectly can produce invalid STAC.
     """
     STACVersion.set_stac_version(stac_version)
